"""
Date

This module contains actions for dealing with dates & times.
"""

import datetime

from typing import TypedDict, Optional

from ppa import ui

TIME_FORMAT = "%H:%M:%S"
DATE_FORMAT = "%d/%m/%Y"
DATE_TIME_FORMAT = f"{TIME_FORMAT} {DATE_FORMAT}"


class Date(TypedDict):
    r"""
- Contains date & time information

- Used as an _output_ from [now](../ppa_tools.date/#now)

!!! example

    ```yaml linenums="1"
    display: Tue Aug 25 08:10:45 2020
    date: 25/08/2020
    time: '08:10:45'
    day: 25
    day_name: Tuesday
    month: 8
    month_name: August
    year: 2020
    hour: 8
    minutes: 10
    seconds: 45
    ```

!!! info "Date Keys"

    __display__: The day, month, date, time, & year displayed together

    __date__: The date in __dd/mm/yyyy__ format

    __time__: The time in __hh:mm:ss__ format

    __day__: The day as a __number__

    __day_name__: The name of the day

    __month__: The month as a __number__

    __month_name__: The name of the month

    __year__: The year as a __number__

    __hour__: The hour as a __number__

    __minutes__: The minutes as a __number__

    __seconds__: The seconds as a __number__

    __timestamp__: The date & time as a timestamp

    """
    display: str
    date: str
    time: str
    day: int
    day_name: str
    month: int
    month_name: str
    year: int
    hour: int
    minutes: int
    seconds: int
    timestamp: int


class Delta(TypedDict):
    total_seconds: int
    total_minutes: int


def _date_to_dict(date_object: datetime.datetime) -> Date:
    return {
        "display": date_object.strftime("%c"),
        "date": date_object.strftime(DATE_FORMAT),
        "time": date_object.strftime(TIME_FORMAT),
        "day": int(date_object.strftime("%d")),
        "day_name": date_object.strftime("%A"),
        "month": int(date_object.strftime("%m")),
        "month_name": date_object.strftime("%B"),
        "year": int(date_object.strftime("%Y")),
        "hour": int(date_object.strftime("%H")),
        "minutes": int(date_object.strftime("%M")),
        "seconds": int(date_object.strftime("%S")),
        "timestamp": int(datetime.datetime.timestamp(date_object)),
    }


def _date_from_dict(date_dict: Date) -> datetime.datetime:
    return datetime.datetime.strptime(f"{date_dict['time']} {date_dict['date']}", DATE_TIME_FORMAT)


def now() -> Date:
    """Get the current ___UTC___ date & time.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        This action takes no inputs.

    ??? success "Output"

        A single [Date](../inputs_outputs/#date) dictionary.

    ??? example

        ```YAML linenums="1"
        - ppa_tools.date.now:
          save: date
        ```
    """
    return _date_to_dict(datetime.datetime.utcnow())


def from_string(string: str) -> Date:
    """Create a [Date](../inputs_outputs/#date) from a date & time string.

The date & time string must be in the following format:

    hour:minute:second day/month/year

For example:

    09:00:00 21/03/2022

!!! tip "All generated dates have a UTC timezome"

!!! question "Minimum Plugin Version: 7.0.0"

??? info "Input"

    __string:__ the date & time string

??? success "Output"

    [Date](../inputs_outputs/#date)

??? example

    ```YAML linenums="1"
    - ppa_tools.date.from_string:
        string: 09:00:00 21/03/2022
      save: date
    ```
    """
    try:
        return _date_to_dict(datetime.datetime.strptime(string, DATE_TIME_FORMAT))
    except ValueError:
        message = f"Supplied date & time '{string}' does not match expected format (hour:minute:second day/month/year)."
        ui.output_error(message)
        raise ValueError(message)


def add_delta(
    days: Optional[int] = 0, hours: Optional[int] = 0, minutes: Optional[int] = 0, start_date: Optional[Date] = None
) -> Date:
    """Create a future [Date](../inputs_outputs/#date) by adding days, hours, & minutes to a date & time.

If no **date** is supplied, the current date & time will be used.

!!! tip "All generated dates have a UTC timezome"

!!! question "Minimum Plugin Version: 7.0.0"

??? info "Input"

    __days:__ number of days to add (defaults to 0)

    __hours:__ number of hours to add (defaults to 0)

    __minutes:__ number of minutes to add (defaults to 0)

    __start_date:__ optional [Date](../inputs_outputs/#date) to use as the start point (defaults to current date & time)

??? success "Output"

    [Date](../inputs_outputs/#date)

??? example

    ```YAML linenums="1"
    - ppa_tools.date.add_delta:
        days: 30
      save: date
    ```
    """
    try:
        return _date_to_dict(
            (datetime.datetime.utcnow() if start_date is None else _date_from_dict(start_date))
            + datetime.timedelta(days=days, hours=hours, minutes=minutes)
        )
    except Exception as e:
        ui.output_error(f"Failed to calculate future date & time stamp: {str(e)}")
        raise e


def get_delta(start: Date, end: Date) -> Delta:
    """Get the total number of seconds between two [Dates](../inputs_outputs/#date).

!!! question "Minimum Plugin Version: 7.0.0"

??? info "Input"

    __start:__ the start [Date](../inputs_outputs/#date)

    __end:__ the end [Date](../inputs_outputs/#date)

??? success "Output"

    A dictionary containing the following keys:

    - **total_seconds**

    - **total_minutes**

    These keys should be treated independently (total_minutes = total_seconds divided by 60).


??? example

    ```YAML linenums="1"
    - ppa_tools.date.now:
      save: now

    - ppa_tools.date.add_delta:
        days: 30
      load:
        start_date: now
      save: future

    - ppa_tools.date.get_delta:
      load:
        start: now
        end: future
      save: delta
    ```
    """
    seconds = int((_date_from_dict(end) - _date_from_dict(start)).total_seconds())
    if seconds < 0:
        message = "The delta action returned a negative number, please make sure the start date is earlier than the end date."
        ui.output_error(message)
        raise ValueError(message)
    return Delta(total_seconds=seconds, total_minutes=int((seconds / 60)))
