"""
Dictionaries

This module contains actions used to process dictionaries.
"""
import logging

from typing import Dict, Union, Any, Optional, TypedDict, List

from ppa import ui


logger = logging.getLogger(__name__)


class DuplicateKey(Exception):
    pass


class InvalidFilter(ValueError):
    pass


class ValueFiltered(Exception):
    pass


class Filter(TypedDict):
    """
- Contains a dictionary key, test value, & filter type.

- Supplied in a list to the [filter_from_list](../ppa_tools.dictionaries/#filter_from_list) dictionary action.

!!! warning "Required Keys"

    - __key__: The dictionary key whose value should be tested.
    - __type__: A filter type (see below for more information).
    - __value__: The value to test against the dictionary value.

!!! tip "Valid Filters"

    All filters & the types of value they support:

    - `equal` (_string_ or _integer_)
    - `not_equal` (_string_ or _integer_)
    - `contains` (_string_)
    - `starts_with` (_string_)
    - `ends_with` (_string_)
    - `greater_than` (_integer_)
    - `less_than` (_integer_)
    - `is` (_null_)
    - `not` (_null_)

    New in version **5.1.0**:

    - `not_contains` (_string_)
    - `not_starts_with` (_string_)
    - `not_ends_with` (_string_)

"""

    key: str
    type: str
    value: Union[str, int, bool, None]


def filter_from_list(dictionaries: List[Dict[str, Any]], filters: List[Filter]) -> List[Dict[str, Any]]:
    """Remove dictionaries from a list by testing the values of one or more keys.

    A dictionary will be filtered if ___any___ of its keys match ___any___ filter.

    !!! question "Minimum Plugin Version: 3.1.0"

    ??? info "Input"

        - __dictionaries:__ a list of dictionaries to filter

        - __filters:__ a list of [Filter](../inputs_outputs/#filter) dictionaries

    ??? success "Output"

        A list of dictionaries from `dictionaries` whose keys matched no filters.

    !!! tip "Case Sensitivity"

        As of plugin version **5.1.0** all text filters will be case-insensitive.

    ??? example

        - Getting all users from Active Directory & saving them as `all_users`

        - Filtering out users whose __mail__ attribute is `null` _or_ __sAMAccountName__ contains `admin`

        - Saving the filtered users as `filtered_users`

        ```YAML linenums="1" hl_lines="4 8 9 10 11 12 13 14"
        - active_directory.users.get_all:
          load:
            domain_controller: domain_controller
          save: all_users

        - ppa_tools.dictionaries.filter_from_list:
            filters:
              - key: mail
                type: is
                value: null
              - key: sAMAccountName
                type: contains
                value: admin
          save: filtered_users
        ```
    """

    logger.info(f"received {len(dictionaries)} dictionaries")

    filter_map = {
        "is": lambda value, test: value is test,
        "not": lambda value, test: value is not test,
        "contains": lambda value, test: test in value,
        "not_contains": lambda value, test: test not in value,
        "equal": lambda value, test: value == test,
        "not_equal": lambda value, test: value != test,
        "greater_than": lambda value, test: value > test,
        "less_than": lambda value, test: value < test,
        "ends_with": lambda value, test: value.endswith(test) if value is not None else False,
        "not_ends_with": lambda value, test: not value.endswith(test) if value is not None else False,
        "not_starts_with": lambda value, test: not value.startswith(test) if value is not None else False,
        "starts_with": lambda value, test: value.startswith(test) if value is not None else False,
    }

    def _test_value(filter_type, supplied_value, test_value) -> bool:
        if all([isinstance(supplied_value, str), isinstance(test_value, str)]):
            supplied_value, test_value = list(map(lambda v: v.lower(), [supplied_value, test_value]))
        try:
            if filter_map[filter_type](supplied_value, test_value):
                raise ValueFiltered()
            return False  # False indicates the value does not match the filter
        except (ValueError, TypeError, AttributeError) as e:
            logger.exception(e)
            raise ValueError(
                f"Failed to run '{filter_type}' test using dictionary value {repr(supplied_value)} and "
                f"test value '{repr(test_value)}'. The values may not compatible with the test."
            )
        except ValueFiltered:
            logger.debug(
                f"'{filter_type}' test returned True using dictionary value {repr(supplied_value)} "
                f"and test value '{repr(test_value)}'"
            )
            return True  # True indicates the value matches the filter

    def _test_dictionary(dictionary) -> bool:
        # Return False to filter the dictionary if every filter test returned True
        for relevant_filter in [f for f in filters if f["key"] in dictionary]:
            if all([_test_value(relevant_filter["type"], dictionary[relevant_filter["key"]], relevant_filter["value"])]):
                return False
        return True

    if [f for f in filters if sorted(list(f.keys())) != sorted(["key", "value", "type"])]:
        raise InvalidFilter("Filter dictionaries must only contain 'key, 'value', & 'type' keys.")

    if invalid_types := list(set([f["type"] for f in filters if f["type"] not in filter_map])):
        raise ValueError(
            "The following filter types are invalid: '{invalid}'. Valid types are: '{valid}'".format(
                invalid=", ".join(invalid_types), valid=", ".join(filter_map.keys())
            )
        )

    results = list(filter(lambda d: _test_dictionary(d), dictionaries))
    logger.debug(f"{len(dictionaries) - len(results)} of {len(dictionaries)} dictionaries were removed")
    return results


def find_in_list(dictionaries: List[Dict[str, Any]], filters: List[Filter]) -> List[Dict[str, Any]]:
    """Extract dictionaries from a list by testing the values of one or more keys.

    A dictionary will be found if ___any___ of its keys match ___any___ filter.

    !!! question "Minimum Plugin Version: 6.1.0"

    ??? info "Input"

        - __dictionaries:__ a list of dictionaries to filter

        - __filters:__ a list of [Filter](../inputs_outputs/#filter) dictionaries

    ??? success "Output"

        A list of dictionaries from `dictionaries` whose keys matched at least one filter.

    !!! tip "Case Sensitivity"

        As of plugin version **5.1.0** all text filters will be case-insensitive.

    ??? example

        - Getting all users from Active Directory & saving them as `all_users`

        - Finding users whose __mail__ attribute is `null`

        - Saving the found users as `users_without_email`

        ```YAML linenums="1" hl_lines="4 8 9 10 11 12 13 14"
        - active_directory.users.get_all:
          load:
            domain_controller: domain_controller
          save: all_users

        - ppa_tools.dictionaries.find_in_list:
            filters:
              - key: mail
                type: is
                value: null
          save: users_without_email
        ```
    """
    return [d for d in dictionaries if d not in filter_from_list(dictionaries, filters)]


def create(**key_values: Any) -> Dict[str, Any]:
    """Create a new dictionary using any number of keys & values.

    !!! question "Minimum Plugin Version: 3.4.0"

    ??? info "Input"

        This action accepts any keys & values the new dictionary should contain.

    !!! tip "Validation"

        Given this action accepts any inputs you supply, PPA cannot perform accurate validation in the Playbook editor.

    ??? success "Output"

        A dictionary containing the supplied keys & values.

    ??? example

        Creating a new dictionary called `connection_info` using both static values & variables:

        ```YAML linenums="1"
        - ppa_tools.dictionaries.create:
            port: 5986
            validate_cert: true
          load:
            address: secrets.ip_address
            username: secrets.username
            password: secrets.password
          save: connection_info
        ```
    """
    return key_values


def insert(
    name: str, value: Union[str, int, bool, float], dictionary: Dict[str, Any], overwrite: Optional[bool] = False
) -> Dict[str, Any]:
    """Add a new key & value to a dictionary.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __name:__ the name of the new key

        - __value:__ the value of the new key

        - __dictionary:__ the dictionary to add the key to

    ??? success "Output"

        The original dictionary with the new key & value inserted.

    ??? example

        Adding a value supplied by the Task Operator to a secrets dictionary:

        ```YAML linenums="1"
        {% raw %}
        - ppa_tools.dictionaries.insert:
            name: ip_address
            value: "{{ ip_address }}"
          load:
            dictionary: vault_secrets
          save: vault_secrets
        {% endraw %}
        ```
    """
    if all([not overwrite, name in dictionary]):
        message = f"Key {name} already exists in the supplied dictionary."
        ui.output_error(message)
        raise DuplicateKey(message)
    dictionary.update(**{name: value})
    return dictionary


def merge(first: Dict[str, Any], second: Dict[str, Any]) -> Dict[str, Any]:
    """Merge two dictionaries into a new dictionary.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __first:__ a dictionary to merge with `second`

        - __second:__ a dictionary to merge with `first`

    ??? success "Output"

        A new dictionary containing the keys & values of from `first` & `second`.

    !!! danger "Duplicate Keys"

        This action will fail if any keys appear in ___both___ `first` & `second`.

    ??? example

        Merging two dictionaries to create a single dictionary containing connection details:

        ```YAML linenums="1"
        {% raw %}
        - ppa_tools.dictionaries.merge:
            first:
              address: 192.168.45.123
              port: 22
            second:
              username: "{{ vault_secrets.username }}"
              password: "{{ vault_secrets.password }}"
          save: connection_details
        {% endraw %}
        ```

        The new `connection_details` variable will contain:

        ```YAML linenums="1"
        address: 192.168.45.123
        port: 22
        username: vault-username  # an example secret value
        password: vault-password  # an example secret value
        ```
    """
    if shared_keys := list(set(first.keys()).intersection(second.keys())):
        message = "Both dictionaries contain the following key(s): '{}'".format(", ".join(shared_keys))
        ui.output_error(message)
        raise DuplicateKey(message)
    return dict(**first, **second)


def get(dictionary: Dict[str, Any], key: str) -> Any:
    """Get the value of the supplied key from a dictionary.

    If the key is missing from the dictionary this action will fail.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __dictionary__: a dictionary to get the value from

        - __key__: the key to get from the dictionary

    ??? success "Output"

        The value of the key inside the dictionary.

    ??? example

        ```YAML linenums="1"
        - active_directory.users.get_interactive:
          load:
            domain_controller: domain_controller
          save: user

        - ppa_tools.dictionaries.get:
            key: objectSid
          load:
            dictionary: user
          save: user_sid
        ```
    """
    try:
        return dictionary[key]
    except KeyError as e:
        ui.output_error(f"Key '{key}' does not exist in the supplied dictionary.")
        raise e


def get_key_values(dictionaries: List[Dict[str, Any]], key: str) -> List[Any]:
    """Get the value of the supplied key from each supplied dictionary.

    If the key is missing from any dictionaries its value will be __null__.

    !!! question "Minimum Plugin Version: 3.7.0"

    ??? info "Input"

        - __dictionaries__: a list of dictionaries to get the values from

        - __key__: the key to get from each dictionary

    ??? success "Output"

        A list of values retrieved from each dictionary using the supplied key.

    !!! tip "Ordering"

        The order of the dictionaries is honoured when getting the values.

        This means the first value in the outputted list will be from the first dictionary, etc.

    ??? example

        Extracting the __sAMAccountName__ from each Active Directory user in a list:

        ```YAML linenums="1"
        - active_directory.users.get_all:
          load:
            domain_controller: domain_controller
          save: all_users

        - ppa_tools.dictionaries.get_key_values:
            key: sAMAccountName
          load:
            dictionaries: all_users
          save: all_samaccountnames
        ```
    """
    return [d.get(key) for d in dictionaries]


def remove_duplicates(items: List[Dict[str, Any]], key: str) -> List[Dict[str, Any]]:
    """Remove all duplicate dictionaries from a list.

    Rather than checking each key in every dictionary, the value of the supplied **key** is used.

    !!! question "Minimum Plugin Version: 5.2.0"

    ??? info "Input"

        - __items:__ a list of dictionaries

        - __key:__ a key present in every dictionary whose value should be used to check for uniqueness

    ??? success "Output"

        A new copy of `items` with any duplicate values removed.

    ??? example

        - Auditing users from 2 Security Groups in Active Directory & combining them into a single list

        - Using the **sAMAcccountName** key to remove duplicate users from the combined list

        ```YAML linenums="1"
        - active_directory.groups.get_users:
            distinguishedName: CN=Group One,DC=Example,DC=Domain
          load:
            domain_controller: domain_controller
          save: group_one

        - active_directory.groups.get_users:
            distinguishedName: CN=Group Two,DC=Example,DC=Domain
          load:
            domain_controller: domain_controller
          save: group_two

        - ppa_tools.lists.combine:
          load:
            first: group_one
            second: group_two
          save: users

        - ppa_tools.lists.remove_duplicates:
            key: sAMAccountName
          load:
            items: users
          save: users
        ```
    """
    seen_values = []
    unique_records = []
    for v, item in enumerate(items):
        if key not in item:
            message = f"One of the supplied dictionaries does not contain the key '{key}'."
            ui.output_error(message)
            raise KeyError(message)
        value = item[key]
        if value not in seen_values:
            seen_values.append(value)
            unique_records.append(item)
    return unique_records
