"""
Generate

This module contains actions used to generate various data.
"""
import random

from uuid import uuid4

from typing import List, Dict, Any, Union, Optional

from ppa import ui


def table_from_dictionaries(
    header: List[str], keys: List[str], dictionaries: List[Dict[str, Any]]
) -> Dict[str, Union[List[str], List[List[str]]]]:
    r"""Generate table data from a list of dictionaries.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __header:__ a list of table column headers

    - __keys:__ a list of corresponding dictionary keys, one for each `header`

    - __dictionaries:__ a list of dictionaries to get the values of each `key` from

??? success "Output"

    A dictionary containing the keys `header` & `rows`.

??? example

    - Finding failed SSH logons using [extract_values](../ppa_tools.text/#extract_values) & saving them as `failed_logons`

    - Using this action to turn `failed_logons` into table data, & saving it as `failed_logons_table`

    - Displaying a table using the data generated by this action

    ```YAML linenums="1" hl_lines="10 20 21 26 27"
    - ppa_tools.text.extract_values:
        extract:
          - name: username
            expression: for (.*) from
          - name: ip_address
            expression: from (\d+\.\d+\.\d+\.\d+) port
        text: >
          2020-08-11T08:19:40.463248+00:00 machine-name sshd[10349]: Failed password for user_1 from 1.2.3.4 port 58041 ssh2
          2020-08-12T13:24:07.463248+00:00 machine-name sshd[10349]: Failed password for user_2 from 2.3.4.5 port 58041 ssh2
      save: failed_logons

    - ppa_tools.generate.table_from_dictionaries:
        header:
          - Username
          - IP Address
        keys:
          - username
          - ip_address
      load:
        dictionaries: failed_logons
      save: failed_logons_table

    - ppa.ui.output_table:
        text: Failed SSH Logons
      load:
        header: failed_logons_table.header
        rows: failed_logons_table.rows
    ```

!!! tip "Headers & Keys"

    The `header` & `keys` lists must be the same length, & the attributes must be in the same order.

    If a supplied key is missing from a dictionary, its value will be blank.
    """

    def _get_value(d, k):
        return d.get(k) if d.get(k) is not None else ""

    if len(header) != len(keys):
        message = "Cannot generate table data as the supplied 'header' & 'keys' lists are not the same length."
        ui.output_error(message)
        raise ValueError(message)
    return {"header": header, "rows": [[_get_value(d, key) for key in keys] for d in dictionaries]}


def html_table(header: List[str], rows: List[List[str]]) -> str:
    r"""Generate an HTML table to use in a markdown document or email.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __header:__ a list of table column headers

    - __rows:__ the table contents as a list of lists, where each inner list is a row of values

??? success "Output"

    The generated HTML table as a string.

??? example

    ```YAML linenums="1"
    - ppa_tools.html.generate_table:
        header:
          - City
          - Country
        rows:
          - - Paris
            - France
          - - London
            - England
      save: html_table
    ```
    """
    header_styles = "; ".join(
        ["padding: 5px", "font-size: 13px", "font-weight: heavy", "text-align: center", "border: 1px solid grey"]
    )
    cell_styles = "; ".join(["padding: 5px", "font-size: 12px", "border: 1px solid gray", "text-align: center"])

    row_template = "<tr>\n{row}\n</tr>"
    table_header = row_template.format(row="\n".join([f'<th style="{header_styles}">{head}</th>' for head in header]))
    table_rows = "\n".join(
        [row_template.format(row="\n".join([f'<td style="{cell_styles}">{cell}</td>' for cell in row])) for row in rows]
    )

    return f"""<table style="margin: auto;">
    {table_header}
    {table_rows}
</table>"""


def random_number(minimum: Optional[int] = 0, maximum: Optional[int] = 10) -> int:
    """Generate a random number within a defined range.

!!! question "Minimum Plugin Version: 8.2.0"

??? info "Input"

    - __minimum:__ the smallest number in the range (defaults to `0`)

    - __maximum:__ the largest number in the range (defaults to `10`)

??? success "Output"

    The generated number.

??? example

    ```YAML linenums="1"
    - ppa_tools.generate.random_number:
      save: number
    ```
    """
    if minimum > maximum:
        message = f"Cannot generate random number between {minimum}-{maximum} (minimum is higher than maximum)."
        ui.output_error(message)
        raise ValueError(message)
    return random.randint(minimum, maximum)


def uuid() -> str:
    """Generate a UUID.

!!! question "Minimum Plugin Version: 8.2.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    The generated UUID as a string.

??? example

    ```YAML linenums="1"
    - ppa_tools.generate.uuid:
      save: uuid
    ```
    """
    return str(uuid4())
