"""
Regex

This module contains actions for searching & modifying text using regular expressions.
"""

import re
import logging


from ppa import ui


logger = logging.getLogger(__name__)


def replace(text: str, expression: str, replacement: str) -> str:
    r"""Replace a string with another inside some text using a regular expression.

    !!! question "Minimum Plugin Version: 5.6.0"

    ??? info "Input"

        - __text:__ the text to process

        - __expression:__ a regular expression that describes the text to replace

        - __replacement:__ a string to replace any expression matches with

    !!! tip "No Matches"

        This action will succeed and return the original __text__ if no expression matches are found.

    ??? success "Output"

        A new copy of __text__ with any replacements made.

    ??? example

        Removing all lines that end with __- England__:

        ```YAML
        - ppa_tools.regex.replace:
            expression: >
              .*\-\sEngland\n
            replacement: ""
            text: >
              London - England
              Manchester - England
              Paris - France
              New York - USA
          save: replaced
        ```

        The output will be:

        ```
        Paris - France
        New York - USA
        ```
    """
    try:
        regex = re.compile(expression)
    except re.error:
        message = f"The supplied regular expression '{expression}' is invalid."
        ui.output_error(message)
        raise ValueError(message)

    if re.search(regex, text) is not None:
        return re.sub(regex, replacement, text)

    logger.info("the supplied expression found no match")
    return text
